/*  GSML - General Sound Manipulation Library
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   ESD Mixer Node implementation (very limited - 16bit and output only)
 */

#include "IO_OSS.hh"

#include <sstream>

using namespace GSML;

/********************************
 * The OSS Node implementation *
 ********************************/

ESDMixerNode::ESDMixerNode ()
{
  initialized = false;
  io_running = false;
  
  do_playback = true;
  do_capture = false;
  
  esd_period_size = 0;
  esd_fd = 0;
  
  Ready (false);
}

ESDMixerNode::~ESDMixerNode ()
{
  if (io_running)
    cout << "ERRORR: BUG: WARNING!!!!~ESDMixerNode: io is running in dtor!?!?!?! ..." << endl;
  
  // TODO ...
}

bool ESDMixerNode::Init ()
{
  if (initialized)
    {
      ExceptInitialized e;
      e.info = "Already Initialized!";
      throw e;
    }
  
  // TODO: add device auto-detection
  
  string interface = "localhost:16001";
  m_interfaces . push_back (interface);
  
  hostname = "localhost";
  port = 16001;
  
  initialized = true;
  
  return initialized;
}

const vector<string>& ESDMixerNode::Interfaces ()
{
  return m_interfaces;
}

const string& ESDMixerNode::Interface ()
{
  return "";
}

bool ESDMixerNode::SetInterface (const string& n_interface)
{
  if (!initialized)
    {
      ExceptNotInitialized e;
      e.info = "Not Initialized!";
      throw e;
    }
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO currently running ...";
      throw e;
    }
  
  // TODO: all config settings ...
  
  return false;
}

track_index ESDMixerNode::PlaybackChannels ()
{
  return esds_channels;
}

track_index ESDMixerNode::CaptureChannels ()
{
  return esd_channels;
}

int ESDMixerNode::PlaybackPeriodSize ()
{
  return esd_period_size;
}

int ESDMixerNode::CapturePeriodSize ()
{
  return esd_period_size;
}

int ESDMixerNode::PlaybackPeriods ()
{
  return esd_periods;
}

int ESDMixerNode::CapturePeriods ()
{
  return esd_periods;
}

int ESDMixerNode::PlaybackXRuns ()
{
  return pl_xruns;
}

int ESDMixerNode::CaptureXRuns ()
{
  return ca_xruns;
}

void ESDMixerNode::SetPlayback (bool state)
{
  if (!initialized)
    {
      ExceptNotInitialized e;
      e.info = "Not Initialized!";
      throw e;
    }
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO currently running ...";
      throw e;
    }
  if (state != true)
    {
      ExceptHWError e;
      e.info = "not yet supported";
      throw e;
    }
  
  do_playback = state;
}

void ESDMixerNode::SetCapture (bool state)
{
  if (!initialized)
    {
      ExceptNotInitialized e;
      e.info = "Not Initialized!";
      throw e;
    }
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO currently running ...";
      throw e;
    }
  if (state != false)
    {
      ExceptHWError e;
      e.info = "not yet supported";
      throw e;
    }
  
  do_capture = state;
}

bool ESDMixerNode::IsPlayback ()
{
  if (!initialized)
    {
      ExceptNotInitialized e;
      e.info = "Not Initialized!";
      throw e;
    }
  
  return do_playback;
}

bool ESDMixerNode::IsCapture ()
{
  if (!initialized)
    {
      ExceptNotInitialized e;
      e.info = "Not Initialized!";
      throw e;
    }
  return do_capture;
}

void ESDMixerNode::StartIO (Mixer& source)
{
  if (!initialized)
    {
      ExceptNotInitialized e;
      e.info = "Not Initialized!";
      throw e;
    }
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO currently running ...";
      throw e;
    }
  m_mixer = &source;
  
  if (do_playback && do_capture)
    oss_fd = open (cname.c_str (), O_RDWR);
  else if (do_playback)
    oss_fd = open (cname.c_str (), O_WRONLY);
  else if (do_capture)
    oss_fd = open (cname.c_str (), O_RDONLY);
  
  if (oss_fd < 0)
    throw ExceptHWError ("ESD - can't open pl");
  
  if (!internal_config_params (false) )
    throw ExceptHWError ("ESD - can't config params pl");
  
  if (do_playback) {
    pl_buffer = new SampleBufferDefault (oss_period_size);
    pl_interleaved = new pcm_16bit::value_type [oss_period_size * oss_channels];
    pl_processed = 0;
    pl_vaild = false;
  }
  
  if (do_capture) {
    ca_buffer = new SampleBufferDefault (oss_period_size);
    ca_interleaved = new pcm_16bit::value_type [oss_period_size * oss_channels];
    ca_processed = 0;
    
    cout << "ca_interleaved: " << ca_interleaved << endl;
  }
  
  // prepare states
  stop_requested = false;
  
  current_playback_period = 0;
  current_capture_period = 0;
  
  pl_xruns = 0;
  ca_xruns = 0;
  
  m_is_first_loop = true;
  m_worked_in_last_loop = true;
  m_using_realtime_sched = false;
  
  io_running = true;
  
#ifdef DEBUG_GSML_IO_ESD
  cout << "  StartIO: calling _Continue_" << endl;
#endif
  Continue (); // unlock Loop ...
}
    
void ESDMixerNode::StopIO ()
{
  if (!io_running)
    {
      ExceptNoIORunning e;
      e.info = "No IO is currently running ...";
      throw e;
    }
  
  stop_requested = true;
}

bool ESDMixerNode::IsIORunning ()
{
  return io_running;
}

sample_index ESDMixerNode::GetCurrentPos ()
{
  if (!io_running)
    {
      ExceptNoIORunning e;
      e.info = "No IO currently running ...";
      throw e;
    }
  
  if (do_playback)
    return current_playback_period * oss_period_size;
  if (do_capture)
    return current_capture_period * oss_period_size;
  
  cout << "Warning: Neither playback nor capture is running!" << endl;
  return 0;
}

void ESDMixerNode::Debug ()
{
  cout << "ESDMixerNode::Debug:" << endl;
  
  cout << "  Period size: " << oss_period_size << endl
       << "  Channels: " << oss_channels << endl;
  
  cout << "  Initialized: " << initialized
       << ", io_running: " << io_running
       << ", do_playback: " << do_playback
       << ", do_capture: " << do_capture << endl;
  
  if (io_running)
    {
      cout << "  stop requested: " << stop_requested << endl;
      
      cout << "  Current period: playback: " << current_playback_period
	   << ", capture: " << current_capture_period << endl;
      
      cout << endl;
    }
}

bool ESDMixerNode::reget_card_params ()
{
  cout << "  ESD: reget for: " << cname << endl;
  if ( (oss_fd = open (cname.c_str (), O_WRONLY) ) < 0)
    return false;
  
  bool ret = internal_config_params (true);
  
  if (close (oss_fd) < 0)
    cout << "  ESD: erro closing file after param config" << endl;
  
  return ret;
}

bool ESDMixerNode::internal_config_params (bool overwrite_params)
{
  int format = AFMT_S16_LE;
  if (ioctl (oss_fd, SNDCTL_DSP_SETFMT, &format) == -1) {
    cout << "  ESD: error setting the sample format" << endl;
    return false;
  }
  if (format != AFMT_S16_LE) {
    cout << "  ESD: format not supported" << endl;
    return false;
  }
  
  if (overwrite_params) {
    m_params.min_depth = 16;
    m_params.max_depth = 16;
    m_params.cur_depth = 16;
  }

  int channels = oss_channels;
  if (ioctl (oss_fd, SNDCTL_DSP_CHANNELS, &channels) == -1) {
    cout << "  ESD: error setting channels" << endl;
    return false;
  }
  if (channels != oss_channels) {
    cout << "  ESD: different channels retuned: " << channels << endl;
    oss_channels = channels;
  }
  
  int rate = m_params.cur_rate;
  if (ioctl (oss_fd, SNDCTL_DSP_SPEED, &rate) == -1) {
    cout << "  ESD: error setting sample rate" << endl;
    return false;
  }
  
  if (rate != m_params.cur_rate) {
    cout << "  ESD: defferent sample rate returned: " << rate << endl;
    m_params.cur_rate = rate;
  }
  
  // some hard config / hack - because it isn't controllable via ESD
  
  if (overwrite_params) {
    m_params.min_rate = 8000;
    m_params.max_rate = 48000;
  }
  
  oss_periods = 8;
  
  if (ioctl (oss_fd, SNDCTL_DSP_GETBLKSIZE, &oss_period_size) == -1) {
    cout << "  ESD: error getting period size" << endl;
    return false;
  }
  oss_period_size /= (2 * oss_channels);
  
  if (overwrite_params) {
    m_params.min_periods = oss_periods;
    m_params.max_periods = oss_periods;
    m_params.cur_periods = oss_periods;
  }
  
  if (overwrite_params) {
    m_params.min_period_time = (1000000 / m_params.cur_rate) * oss_period_size;
    m_params.max_period_time = (1000000 / m_params.cur_rate) * oss_period_size;
    m_params.cur_period_time = (1000000 / m_params.cur_rate) * oss_period_size;
  }
  
  cout << "  ESD: using: "
       << "channels: " << oss_channels
       << ", rate: " << m_params.cur_rate
       << ", periods: " << oss_periods
       << ", period_size: " << oss_period_size
       << ", period_time: " << m_params.cur_period_time
       << endl;
  
  return true;
}

#include <errno.h>

// the second thread ...
void ESDMixerNode::Loop ()
{
  bool data = true;
  
  if (m_is_first_loop) // do first loop things (real-time scheduling ...)
    {
      m_using_realtime_sched = ThreadHelper::EnableRealtimeScheduling ();
      if (!m_using_realtime_sched)
	cout << "Warning: ESDNode: Can't enable realtime scheduling ..." << endl;
      m_is_first_loop = false;
    }
  
#ifdef DEBUG_GSML_IO_ESD
  cout << "ESDMixerNode::Loop:" << endl;
#endif
  try {
  if (!stop_requested)
    {
      // start capture ??
      if (do_capture && !cap_running)
	{
	  /* TODO: port to ESD ??
	  cout << "capture is not running ... -> starting now." << endl;
	  if (!capture_card->Start () )
	    cout << "can't start capture!" << endl;
	  else
	  */
	  cap_running = true;
	}
      
      // sleep when not worked in last loop
      if (!m_worked_in_last_loop)
	{
	  // TODO: to a real computation
	  int xx_delay = 3000;
	  
	  // make shoure other tasks get time ...
	  if (m_using_realtime_sched)
	    ThreadHelper::USleep (xx_delay);
	  else
	    usleep (xx_delay);
	}
      
      m_worked_in_last_loop = false;
      
      if (do_playback)
	{
	  if (!pl_vaild) {
	    // read from mixer and build interleaved buffer ...
	    for (track_index ch = 0; ch < oss_channels; ++ ch)
	      {
		pl_buffer -> Clear ();
		bool temp = m_mixer -> Read (ch, pl_buffer);
		data = data && temp;
		
		for (int i = 0; i < oss_period_size; ++ i)
		  pl_interleaved [i * oss_channels + ch]
		    ( pcm_24_8bit::Convert2int16 (pl_buffer -> data [i] ) );
	      }
	    pl_vaild = true;
	    pl_processed = 0;
	  }
	  
	  /* DEBUG code
	     audio_buf_info t_binfo;
	     if (ioctl (oss_fd, SNDCTL_DSP_GETOSPACE, &t_binfo) == -1)
	     cout << "  ESD: playback error getting buffer into" << endl;
	     else {
	     cout << "  ESD: playback buffer: fragments: " << t_binfo.fragments
	     << ", frags_total: " << t_binfo.fragstotal
	     << ", fragsize: " << t_binfo.fragsize
	     << ", bytes: " << t_binfo.bytes
	     << endl;
	     }
	  */
	  
	  // the select stuff doesn't work ??
	  //if (IsWriteable (oss_fd) )
	  //  {
	      int t_period_samples = oss_period_size * oss_channels;
	      int t_written = write (oss_fd,
				     pl_interleaved + pl_processed,
				     (t_period_samples - pl_processed) * 2); // gets bytes
	      
	      if (t_written > 0) {
		if (t_written % 2 != 0)
		  cout << "ESD: BUG: WRITE ACCEPTED A NON MULTIPLE OF 2??" << endl;
		
		pl_processed += (t_written / 2);
		
		// period-buffer completely processed ??
		if (pl_processed >= t_period_samples) {
		  if (pl_processed > t_period_samples)
		    cout << "ESD: BUG: wrote more samples than period size??" << endl;
		  
		  pl_vaild = false;
		  ++ current_playback_period;
		}
		m_worked_in_last_loop = true;
	      }
	      else
		cout << "ESD: write error: " << t_written << endl;
	      // }
	      // else
	      // cout << "file not writeable ..." << endl;
	} // end if playback
      
      
      if (do_capture)
	{
	  /* DEBUG code
	     audio_buf_info t_binfo;
	     if (ioctl (oss_fd, SNDCTL_DSP_GETISPACE, &t_binfo) == -1)
	     cout << "  ESD: capture error getting buffer into" << endl;
	     else {
	     cout << "  ESD: capture buffer: fragments: " << t_binfo.fragments
	     << ", frags_total: " << t_binfo.fragstotal
	     << ", fragsize: " << t_binfo.fragsize
	     << ", bytes: " << t_binfo.bytes
	     << endl;
	     }
	  */

	  int t_period_samples = oss_period_size * oss_channels;
	  int t_needed = t_period_samples - ca_processed;
	  
	  //cout << "ca_processed: " << ca_processed
	  //     << ", t_period_samples: " << t_period_samples
	  //     << ", t_needed: " << t_needed << endl;
	  
	  int t_read = read (oss_fd,
			     ca_interleaved + ca_processed,
			     t_needed * 2);
	  
	  if (t_read > 0) {
	    if (t_read % 2 != 0)
	      cout << "ESD: BUG: READ RETUNED A NON MULTIPLE OF 2??" << endl;
	    
	    ca_processed += (t_read / 2);
	    
	    // is the period-buffer completey filled ??
	    if (ca_processed >= t_period_samples)
	      {
		if (ca_processed > t_period_samples)
		  cout << "ESD: BUG: processed more samples than period size??" << endl;
		
		ca_buffer -> SetUsedLength (oss_period_size);
		
		// deinterleave and write to mixer
		for (track_index ch = 0; ch < oss_channels; ++ ch)
		  {
		    for (int i = 0; i < oss_period_size; ++ i)
		      ca_buffer -> data [i] ( ca_interleaved [i * oss_channels + ch] << 8 );
		    
		    bool temp = m_mixer -> Write (ch, ca_buffer);
		    data = data && temp;
		  }
		
    		++ current_capture_period;
		ca_processed = 0;
	      }
	    m_worked_in_last_loop = true;
	  }
	  else {
	    cout << "ESD: read error: " << t_read << endl;
	    cout << "     error: " << errno << ": " << strerror (errno) << endl;
	  }
	} // end if capture
    }
  
  // no else - because it might be changed in the io loop-part ...
  if (!data || stop_requested)
    {
      //#ifdef DEBUG_GSML_IO_ALSA
      cout << "  Loop: END OF IO" << endl;
      //#endif
      
      /* clear some states and data */
      
      cout << "closind cards ..." << endl;
      close (oss_fd);
      
      cout << "deleteing buffers ..." << endl;
      
      if (do_playback) {
	delete pl_buffer;
	delete [] pl_interleaved;
      }
      
      if (do_capture) {
	delete ca_buffer;
	delete [] ca_interleaved;
      }
      
      io_running = false;
      m_mixer->IOFinishedNotification ();
      
      Wait ();
    }
  }
  catch (GSML::BasicExcept e)
    {
      cout << endl 
	   << "!! ESDMixerNode::Loop caught exception: !!" << endl
	   << e.info << endl;
      throw;
    }
}
