/* Factory.cc  (C) by V. Ziegler 2001
 
 ***************************************************************************
 *                                                                         *
 *   This code is free software; you can redistribute it and / or modify   *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "Factory.hh"
#include "PluginInterface.hh"
#include <map>
#include <dlfcn.h>

namespace GSMX_utils
{
  namespace Plugin {

    // depends on compiler. TODO: auto configure
    const string dl_prefix = "";
    const string dl_postfix = "";
    // symbol of info function
    const string dl_get_interface_funct = "GSML_plugin_get_interface";

    class plugin
    {
    public:
      string dl_filename;
      void* dl_handle;
      void* dl_symbol;
      Interface* interface;
      plugin (string filename, Interface* iptr) : dl_filename (filename), interface (iptr)
      {
	dl_handle = 0;
	dl_symbol = 0;
      };
    };


    class Internals
    {
    public:
      static map <string, map <string, plugin*> > plugins;

      static map <string, void*> dl_handles;    // handles for all open .so files 
      static map <string, void*> dl_symbols;    // pointer to previously resolved symbols

      static map <string, Info*> aliases;

      static DlFileList dl_names;           // List of all registered dl files
    };

    extern map <string, map <string, plugin*> > Internals::plugins;
    extern map <string, void*> Internals::dl_handles;
    extern map <string, void*> Internals::dl_symbols;
    extern map <string, Info*> Internals::aliases;
    extern DlFileList Internals::dl_names;


    /* forward */
    bool CheckExistence (const string& category, const string& name, bool throw_exception);





    void FuckUp () 
    {
      cout << "in App: " << Internals::plugins.size () << endl;

      cout << "X PLUGINS PTR " << &Internals::plugins << endl;
 
      for (map <string, map <string, plugin*> > :: const_iterator j = Internals::plugins.begin (); j != Internals::plugins.end (); j++)
	{
	  cout << j -> first << endl;
	}
    }







    /************* plugin system API ***************/

    pluginset* RegisterFile (const string& filename)
    {
      cout << "Registering file " << filename << endl;

      cout << "before dlopen: " << Internals::plugins.size () << endl;
      for (map <string, map <string, plugin*> > :: const_iterator j = Internals::plugins.begin (); j != Internals::plugins.end (); j++)
	{
	  cout << j -> first << endl;
	}
      cout << "X PLUGINS PTR " << &Internals::plugins << endl;

      void* handle = dlopen (filename.c_str (), RTLD_NOW);
      if (! handle)
	throw ExceptionOpenFailed (dlerror ());

      cout << "after dlopen: " << Internals::plugins.size () << endl;
      for (map <string, map <string, plugin*> > :: const_iterator j = Internals::plugins.begin (); j != Internals::plugins.end (); j++)
	{
	  cout << j -> first << endl;
	}
      cout << "X PLUGINS PTR " << &Internals::plugins << endl;

 
      string interface_funct = dl_prefix
	+ dl_get_interface_funct
	+ dl_postfix;

      get_interface ifunc = (get_interface) dlsym (handle, interface_funct.c_str ());
      if (!ifunc)
	throw ExceptionNoPluginDl (filename);

      const deque <Interface*> new_plugins = ifunc ();
 
      pluginset* just_registered = new pluginset;

      cout << new_plugins . size () << " Plugins" << endl;

      for (deque <Interface*>::const_iterator i = new_plugins.begin (); i != new_plugins.end (); ++i)
	{
	  Interface* current = *i;
	  cout << "Registering Plugin " << current -> info . category << " : " << current -> info . name << endl;
	  cout << "by " << current -> info . author << " (" << current -> info . date << ")" << endl;
	  cout << current -> constructor << endl;

	  /* TODO: on name conflict, clean up and unregister everything from causing file !!!! */
	  plugin* entry = new plugin (filename, current);
	  entry -> dl_handle = handle;

	  /* Check for name conflicts..... */
	  map <string, map <string, plugin*> > :: const_iterator conflict = Internals::plugins.find (current -> info . category); 
	  if (conflict != Internals::plugins.end ())
	    if (conflict -> second . find (current -> info . name) != conflict -> second . end ())
	      throw ExceptionNameConflict (current -> info . category, current -> info . name);

	  Internals::plugins [current -> info . category] [current -> info . name] = entry;

	  cout << "after assignment: " << Internals::plugins.size () << endl;
	  for (map <string, map <string, plugin*> > :: const_iterator j = Internals::plugins.begin (); j != Internals::plugins.end (); j++)
	    {
	      cout << j -> first << endl;
	    }
	  cout << "X PLUGINS PTR " << &Internals::plugins << endl;

	  just_registered -> push_back (& (entry -> interface -> info));


	  cout << "after assignment 2: " << Internals::plugins.size () << endl;
	  for (map <string, map <string, plugin*> > :: const_iterator j = Internals::plugins.begin (); j != Internals::plugins.end (); j++)
	    {
	      cout << j -> first << endl;
	    }
	  cout << "X PLUGINS PTR " << &Internals::plugins << endl;

	}
      cout << "Finished, closing dl" << endl;

      cout << "before dlclose: " << Internals::plugins.size () << endl;
      for (map <string, map <string, plugin*> > :: const_iterator j = Internals::plugins.begin (); j != Internals::plugins.end (); j++)
	{
	  cout << j -> first << endl;
	}
      cout << "X PLUGINS PTR " << &Internals::plugins << endl;

      //dlclose (handle);

      cout << "after dlclose: " << Internals::plugins.size () << endl;
      for (map <string, map <string, plugin*> > :: const_iterator j = Internals::plugins.begin (); j != Internals::plugins.end (); j++)
	{
	  cout << j -> first << endl;
	}
      cout << "X PLUGINS PTR " << &Internals::plugins << endl;

      cout << "dl closed." << endl;

      Internals::dl_names . push_back (filename); 

      cout << "before return: " << Internals::plugins.size () << endl;
      for (map <string, map <string, plugin*> > :: const_iterator j = Internals::plugins.begin (); j != Internals::plugins.end (); j++)
	{
	  cout << j -> first << endl;
	}
      cout << "X PLUGINS PTR " << &Internals::plugins << endl;


      return just_registered;
    }

    content* Categorys ()
    {
      content* ret = new content;
      for (map<string, map <string, plugin*> >::const_iterator i = Internals::plugins.begin (); i != Internals::plugins.end (); ++i)
	ret -> push_back (string (i -> first));	
      return ret;
    }

    content* Items (const string& category)
    {
      /* check for existence */
      if (Internals::plugins.find (category) == Internals::plugins.end ())
	throw ExceptionUnknownCategory (category);

      content* ret = new content;
      for (map <string, plugin*>::const_iterator i = Internals::plugins [category].begin (); i != Internals::plugins [category].end (); ++i)
	ret -> push_back (string (i -> first));	
      return ret;
    }


    const Info& PlugInfo (const string& category, const string& name)
    {
      /* check for existence */
      CheckExistence (category, name, true);
      return Internals::plugins [category] [name] -> interface -> info;
    }

    extern void* __LocateGenerator (const string& category, const string& name)
    {                   
      /* check for existence */
      CheckExistence (category, name, true);

      plugin* requested = Internals::plugins [category][name];
 
      if (!requested -> dl_handle)
	{
	  void* handle = Internals::dl_handles [requested -> dl_filename];
	  if (!handle)
	    {
	      cout << "opening " << requested -> dl_filename << endl;
	      handle = dlopen (requested -> dl_filename.c_str (),  RTLD_NOW);
	      if (! handle)
		throw ExceptionOpenFailed (dlerror ());
	     
	      Internals::dl_handles [requested -> dl_filename] = handle;
	    }
	  requested -> dl_handle = handle;
	}

      if  (!requested -> dl_symbol)
	{
	  void* symbol = Internals::dl_symbols [requested -> interface -> constructor];
	  if (!symbol)
	    {
	      cout << "locating generator " << requested -> interface -> constructor << endl;
	      string interface_funct = dl_prefix
		+ requested -> interface -> constructor
		+ dl_postfix;

	      symbol = dlsym (requested -> dl_handle, interface_funct.c_str ());
	      if (!symbol )
		throw ExceptionDlError (dlerror ());

	      Internals::dl_symbols [requested -> interface -> constructor] = symbol;
	    }
	  requested -> dl_symbol = symbol;
	}
    
      return requested -> dl_symbol;
    }

    const DlFileList& RegisteredDlFiles ()
    {
      return Internals::dl_names;
    }


    void Alias (const string& category, const string& name, const string& alias)
    {
      /* check for existence */
      CheckExistence (category, name, true);

      Info* info = &(Internals::plugins [category] [name] -> interface -> info);
      Internals::aliases [alias] = info;      
    }

    const Info& ResolveAlias (const string& alias)
    {
      // TODO: check for existence !!!!!!
      return *(Internals::aliases [alias]);
    }

    void UnAlias (const string& alias)
    {
      Internals::aliases.erase (alias);
    }


    // This Function Checks for existence of a Plugin
    bool CheckExistence (const string& category, const string& name, bool throw_exception)
    {
      for (map <string, map <string, plugin*> > :: const_iterator j = Internals::plugins.begin (); j != Internals::plugins.end (); j++)
	{
	  cout << j -> first << endl;
	}

      cout << "X PLUGINS PTR " << &Internals::plugins << endl;


      map <string, map <string, plugin*> > :: const_iterator i = Internals::plugins.find (category);
      if (i == Internals::plugins.end ())
	{
	  if (throw_exception)
	    throw ExceptionUnknownCategory (category);
	  return false;
	}

      if (i -> second . find (name) == i -> second . end ())
	{
	  if (throw_exception)
	    throw ExceptionUnknownName (category, name);
	  return false;
	}

      return true;
    }

  }; // namespace Plugin
} // namespace GSMX_utils
